# Copyright 2016 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

HTTP_METHODS: list[str] = ['POST', 'PUT', 'GET', 'DELETE', 'PATCH']

BOOT_DEVICES: list[str] = ['pxe', 'disk', 'cdrom', 'bios', 'safe', 'wanboot']

# Polling intervals in seconds.
_LONG_ACTION_POLL_INTERVAL: int = 10
_SHORT_ACTION_POLL_INTERVAL: int = 2
# This dict acts as both list of possible provision actions and arguments for
# wait_for_provision_state invocation.
PROVISION_ACTIONS: dict[str, dict[str, str | int] | None] = {
    'active': {'expected_state': 'active',
               'poll_interval': _LONG_ACTION_POLL_INTERVAL},
    'deleted': {'expected_state': 'available',
                'poll_interval': _LONG_ACTION_POLL_INTERVAL},
    'rebuild': {'expected_state': 'active',
                'poll_interval': _LONG_ACTION_POLL_INTERVAL},
    'inspect': {'expected_state': 'manageable',
                # This is suboptimal for in-band inspection, but it's probably
                # not worth making people wait 10 seconds for OOB inspection
                'poll_interval': _SHORT_ACTION_POLL_INTERVAL},
    'provide': {'expected_state': 'available',
                # This assumes cleaning is in place
                'poll_interval': _LONG_ACTION_POLL_INTERVAL},
    'manage': {'expected_state': 'manageable',
               'poll_interval': _SHORT_ACTION_POLL_INTERVAL},
    'clean': {'expected_state': 'manageable',
              'poll_interval': _LONG_ACTION_POLL_INTERVAL},
    'adopt': {'expected_state': 'active',
              'poll_interval': _SHORT_ACTION_POLL_INTERVAL},
    'abort': None,  # no support for --wait in abort
    'rescue': {'expected_state': 'rescue',
               'poll_interval': _LONG_ACTION_POLL_INTERVAL},
    'unrescue': {'expected_state': 'active',
                 'poll_interval': _LONG_ACTION_POLL_INTERVAL},
    'service': {'expected_state': 'active',
                'poll_interval': _LONG_ACTION_POLL_INTERVAL},
}

PROVISION_STATES: list[str] = list(PROVISION_ACTIONS)



#
#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
import itertools
import logging
from typing import Any

from osc_lib import utils as oscutils

from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc
from ironicclient.osc import command
from ironicclient.v1 import resource_fields as res_fields


class ListBaremetalConductor(command.Lister):
    """List baremetal conductors"""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalConductor")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of conductors to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Baremetal API Service.')
        )
        parser.add_argument(
            '--marker',
            metavar='<conductor>',
            help=_('Hostname of the conductor (for example, of the last '
                   'conductor in the list from a previous request). Returns '
                   'the list of conductors after this conductor.')
        )
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            help=_('Sort output by specified conductor fields and directions '
                   '(asc or desc) (default: asc). Multiple fields and '
                   'directions can be specified, separated by comma.'),
        )
        display_group = parser.add_mutually_exclusive_group(required=False)
        display_group.add_argument(
            '--long',
            default=False,
            help=_("Show detailed information about the conductors."),
            action='store_true')
        display_group.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.CONDUCTOR_DETAILED_RESOURCE.fields,
            help=_("One or more conductor fields. Only these fields will be "
                   "fetched from the server. Can not be used when '--long' "
                   "is specified."))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        client = manager.baremetal

        columns = res_fields.CONDUCTOR_RESOURCE.fields

        params: dict[str, object] = {}
        if parsed_args.limit is not None and parsed_args.limit < 0:
            raise exc.CommandError(
                _('Expected non-negative --limit, got %s') %
                parsed_args.limit)
        params['limit'] = parsed_args.limit
        params['marker'] = parsed_args.marker
        if parsed_args.long:
            params['detail'] = parsed_args.long
            columns = res_fields.CONDUCTOR_DETAILED_RESOURCE.fields
        elif parsed_args.fields:
            params['detail'] = False
            fields = itertools.chain.from_iterable(parsed_args.fields)
            resource = res_fields.Resource(list(fields))
            columns = resource.fields
            params['fields'] = columns

        self.log.debug("params(%s)", params)
        data = client.conductor.list(**params)

        data = oscutils.sort_items(data, parsed_args.sort)

        return (columns,
                (oscutils.get_item_properties(s, columns, formatters={
                    'Properties': utils.HashColumn},) for s in data))


class ShowBaremetalConductor(command.ShowOne):
    """Show baremetal conductor details"""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ShowBaremetalConductor")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "conductor",
            metavar="<conductor>",
            help=_("Hostname of the conductor"))
        parser.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            choices=res_fields.CONDUCTOR_DETAILED_RESOURCE.fields,
            default=[],
            help=_("One or more conductor fields. Only these fields will be "
                   "fetched from the server."))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal
        fields: list[str] | None
        fields = (
            list(itertools.chain.from_iterable(parsed_args.fields))
            or None)
        conductor = baremetal_client.conductor.get(
            parsed_args.conductor, fields=fields)._info
        conductor.pop("links", None)

        return self.dict2columns(conductor)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from collections.abc import Callable
import logging
from typing import Any

from ironicclient.common import base
from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc


LOG: logging.Logger = logging.getLogger(__name__)


class Allocation(base.Resource):
    def __repr__(self) -> str:
        return "<Allocation %s>" % self._info


class AllocationManager(base.CreateManager[Allocation]):
    resource_class: type[Allocation] = Allocation
    _resource_name: str = 'allocations'
    _creation_attributes: list[str] = [
        'extra', 'name', 'resource_class', 'uuid',
        'traits', 'candidate_nodes', 'node', 'owner',
    ]

    def get(
        self,
        allocation_id: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource | None:
        """Get an allocation with the specified identifier.

        :param allocation_id: The UUID or name of an allocation.
        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: an :class:`Allocation` object.

        """
        return self._get(
            resource_id=allocation_id,
            fields=fields,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def delete(
        self,
        allocation_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Delete the Allocation.

        :param allocation_id: The UUID or name of an allocation.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        return self._delete(
            resource_id=allocation_id,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def wait(
        self,
        allocation_id: str,
        timeout: int | float = 0,
        poll_interval: int | float = 1,
        poll_delay_function: (
            Callable[[int | float], object] | None
        ) = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource | None:
        """Wait for the Allocation to become active.

        :param timeout: timeout in seconds, no timeout if 0.
        :param poll_interval: interval in seconds between polls.
        :param poll_delay_function: function to use to wait between polls
            (defaults to time.sleep). Should take one argument - delay time
            in seconds. Any exceptions raised inside it will abort the wait.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :return: updated :class:`Allocation` object.
        :raises: StateTransitionFailed if allocation reaches the error state.
        :raises: StateTransitionTimeout on timeout.
        """
        timeout_msg = _(
            'Allocation %(allocation)s failed to become active '
            'in %(timeout)s seconds') % {
                'allocation': allocation_id,
                'timeout': timeout}
        for _count in utils.poll(
            timeout, poll_interval, poll_delay_function,
            timeout_msg,
        ):
            allocation = self.get(
                allocation_id,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
            if allocation is None:
                continue
            if allocation.state == 'error':
                raise exc.StateTransitionFailed(
                    _('Allocation %(allocation)s failed: '
                      '%(error)s') %
                    {'allocation': allocation_id,
                     'error': allocation.last_error})
            elif allocation.state == 'active':
                return allocation

            LOG.debug(
                'Still waiting for allocation %(allocation)s'
                ' to become active, the current state is'
                ' %(actual)s',
                {'allocation': allocation_id,
                 'actual': allocation.state})
        return None

    def update(
        self,
        allocation_id: str,
        patch: list[dict[str, Any]],
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource:
        """Updates the Allocation. Only 'name' and 'extra' field are allowed.

        :param allocation_id: The UUID or name of an allocation.
        :param patch: a json PATCH document to apply to this allocation.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        return self._update(
            resource_id=allocation_id,
            patch=patch,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def list(
        self,
        resource_class: str | None = None,
        state: str | None = None,
        node: str | None = None,
        limit: int | None = None,
        marker: str | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        fields: list[str] | None = None,
        owner: str | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Allocation]:
        """Retrieve a list of allocations.

        :param resource_class: Optional, get allocations with this resource
                               class.
        :param state: Optional, get allocations in this state. One of
                      ``allocating``, ``active`` or ``error``.
        :param node: UUID or name of the node of the allocation.
        :param marker: Optional, the UUID of an allocation, eg the last
                       allocation from a previous result set. Return
                       the next result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of allocations to return.
            2) limit == 0, return the entire list of allocations.
            3) limit == None, the number of items returned respect the
               maximum imposed by the Ironic API (see Ironic's
               api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned.
        :param owner: Optional, project that owns the allocation.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of allocations.
        :raises: InvalidAttribute if a subset of fields is requested with
                 detail option set.

        """
        if limit is not None:
            limit = int(limit)

        filters = utils.common_filters(
            marker, limit, sort_key, sort_dir, fields)
        for name, value in [('resource_class', resource_class),
                            ('state', state), ('node', node),
                            ('owner', owner)]:
            if value is not None:
                filters.append('%s=%s' % (name, value))

        if filters:
            path = '?' + '&'.join(filters)
        else:
            path = ''
        if limit is None:
            return self._list(
                self._path(path),
                "allocations",
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path),
                "allocations",
                limit=limit,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
import itertools
import logging
from typing import Any

from osc_lib import utils as oscutils

from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc
from ironicclient.osc import command
from ironicclient.v1 import resource_fields as res_fields


class CreateBaremetalVolumeTarget(command.ShowOne):
    """Create a new baremetal volume target."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".CreateBaremetalVolumeTarget")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--node',
            dest='node_uuid',
            metavar='<uuid>',
            required=True,
            help=_('UUID of the node that this volume target belongs to.'))
        parser.add_argument(
            '--type',
            dest='volume_type',
            metavar="<volume type>",
            required=True,
            help=_("Type of the volume target, e.g. 'iscsi', "
                   "'fibre_channel'."))
        parser.add_argument(
            '--property',
            dest='properties',
            metavar="<key=value>",
            action='append',
            help=_("Key/value property related to the type of this volume "
                   "target. Can be specified multiple times."
                   ))
        parser.add_argument(
            '--boot-index',
            dest='boot_index',
            metavar="<boot index>",
            type=int,
            required=True,
            help=_("Boot index of the volume target."))
        parser.add_argument(
            '--volume-id',
            dest='volume_id',
            metavar="<volume id>",
            required=True,
            help=_("ID of the volume associated with this target."))
        parser.add_argument(
            '--uuid',
            dest='uuid',
            metavar='<uuid>',
            help=_("UUID of the volume target."))
        parser.add_argument(
            '--extra',
            dest='extra',
            metavar="<key=value>",
            action='append',
            help=_("Record arbitrary key/value metadata. "
                   "Can be specified multiple times."))

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)" % parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        if parsed_args.boot_index < 0:
            raise exc.CommandError(
                _('Expected non-negative --boot-index, got %s') %
                parsed_args.boot_index)

        field_list: list[str] = [
            'extra', 'volume_type', 'properties',
            'boot_index', 'node_uuid', 'volume_id', 'uuid']
        fields = dict((k, v) for (k, v) in vars(parsed_args).items()
                      if k in field_list and v is not None)
        fields = utils.args_array_to_dict(fields, 'properties')
        fields = utils.args_array_to_dict(fields, 'extra')
        volume_target = baremetal_client.volume_target.create(**fields)

        data = dict([(f, getattr(volume_target, f, '')) for f in
                     res_fields.VOLUME_TARGET_DETAILED_RESOURCE.fields])
        return self.dict2columns(data)


class ShowBaremetalVolumeTarget(command.ShowOne):
    """Show baremetal volume target details."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ShowBaremetalVolumeTarget")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'volume_target',
            metavar='<id>',
            help=_("UUID of the volume target."))
        parser.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.VOLUME_TARGET_DETAILED_RESOURCE.fields,
            help=_("One or more volume target fields. Only these fields will "
                   "be fetched from the server."))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal
        fields: list[str] | None
        fields = (
            list(itertools.chain.from_iterable(parsed_args.fields))
            or None)

        volume_target = baremetal_client.volume_target.get(
            parsed_args.volume_target, fields=fields)._info

        volume_target.pop("links", None)
        columns, data = zip(*sorted(volume_target.items()))
        return columns, data


class ListBaremetalVolumeTarget(command.Lister):
    """List baremetal volume targets."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalVolumeTarget")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--node',
            dest='node',
            metavar='<node>',
            help=_("Only list volume targets of this node (name or UUID)."))
        parser.add_argument(
            '--limit',
            dest='limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of volume targets to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Baremetal API Service.'))
        parser.add_argument(
            '--marker',
            dest='marker',
            metavar='<volume target>',
            help=_('Volume target UUID (for example, of the last '
                   'volume target in the list from a previous request). '
                   'Returns the list of volume targets after this UUID.'))
        parser.add_argument(
            '--sort',
            dest='sort',
            metavar='<key>[:<direction>]',
            help=_('Sort output by specified volume target fields and '
                   'directions (asc or desc) (default:asc). Multiple fields '
                   'and directions can be specified, separated by comma.'))

        display_group = parser.add_mutually_exclusive_group(required=False)
        display_group.add_argument(
            '--long',
            dest='detail',
            action='store_true',
            default=False,
            help=_("Show detailed information about volume targets."))
        display_group.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.VOLUME_TARGET_DETAILED_RESOURCE.fields,
            help=_("One or more volume target fields. Only these fields will "
                   "be fetched from the server. Can not be used when "
                   "'--long' is specified."))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)" % parsed_args)
        manager = self.app.client_manager
        client = manager.baremetal

        columns = res_fields.VOLUME_TARGET_RESOURCE.fields

        params: dict[str, object] = {}
        if parsed_args.limit is not None and parsed_args.limit < 0:
            raise exc.CommandError(
                _('Expected non-negative --limit, got %s') %
                parsed_args.limit)
        params['limit'] = parsed_args.limit
        params['marker'] = parsed_args.marker
        if parsed_args.node is not None:
            params['node'] = parsed_args.node

        if parsed_args.detail:
            params['detail'] = parsed_args.detail
            columns = res_fields.VOLUME_TARGET_DETAILED_RESOURCE.fields
        elif parsed_args.fields:
            params['detail'] = False
            fields = itertools.chain.from_iterable(parsed_args.fields)
            resource = res_fields.Resource(list(fields))
            columns = resource.fields
            params['fields'] = columns

        self.log.debug("params(%s)" % params)
        data = client.volume_target.list(**params)

        data = oscutils.sort_items(data, parsed_args.sort)

        return (columns,
                (oscutils.get_item_properties(s, columns, formatters={
                    'Properties': utils.HashColumn},) for s in data))


class DeleteBaremetalVolumeTarget(command.Command):
    """Unregister baremetal volume target(s)."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".DeleteBaremetalVolumeTarget")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'volume_targets',
            metavar='<volume target>',
            nargs='+',
            help=_("UUID(s) of the volume target(s) to delete."))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        failures: list[str] = []
        for volume_target in parsed_args.volume_targets:
            try:
                baremetal_client.volume_target.delete(volume_target)
                print(_('Deleted volume target %s') % volume_target)
            except exc.ClientException as e:
                failures.append(_("Failed to delete volume target "
                                  "%(volume_target)s: %(error)s")
                                % {'volume_target': volume_target,
                                   'error': e})

        if failures:
            raise exc.ClientException("\n".join(failures))


class SetBaremetalVolumeTarget(command.Command):
    """Set baremetal volume target properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".SetBaremetalVolumeTarget")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'volume_target',
            metavar='<volume target>',
            help=_("UUID of the volume target."))
        parser.add_argument(
            '--node',
            dest='node_uuid',
            metavar='<uuid>',
            help=_('UUID of the node that this volume target belongs to.'))
        parser.add_argument(
            '--type',
            dest='volume_type',
            metavar="<volume type>",
            help=_("Type of the volume target, e.g. 'iscsi', "
                   "'fibre_channel'."))
        parser.add_argument(
            '--property',
            dest='properties',
            metavar="<key=value>",
            action='append',
            help=_("Key/value property related to the type of this volume "
                   "target. Can be specified multiple times."))
        parser.add_argument(
            '--boot-index',
            dest='boot_index',
            metavar="<boot index>",
            type=int,
            help=_("Boot index of the volume target."))
        parser.add_argument(
            '--volume-id',
            dest='volume_id',
            metavar="<volume id>",
            help=_("ID of the volume associated with this target."))
        parser.add_argument(
            '--extra',
            dest='extra',
            metavar="<key=value>",
            action='append',
            help=_("Record arbitrary key/value metadata. "
                   "Can be specified multiple times."))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        if parsed_args.boot_index is not None and parsed_args.boot_index < 0:
            raise exc.CommandError(
                _('Expected non-negative --boot-index, got %s') %
                parsed_args.boot_index)

        properties: list[dict[str, Any]] = []
        if parsed_args.node_uuid:
            properties.extend(utils.args_array_to_patch(
                'add', ["node_uuid=%s" % parsed_args.node_uuid]))
        if parsed_args.volume_type:
            properties.extend(utils.args_array_to_patch(
                'add', ["volume_type=%s" % parsed_args.volume_type]))
        if parsed_args.boot_index:
            properties.extend(utils.args_array_to_patch(
                'add', ["boot_index=%s" % parsed_args.boot_index]))
        if parsed_args.volume_id:
            properties.extend(utils.args_array_to_patch(
                'add', ["volume_id=%s" % parsed_args.volume_id]))

        if parsed_args.properties:
            properties.extend(utils.args_array_to_patch(
                'add', ["properties/" + x for x in parsed_args.properties]))
        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch(
                'add', ["extra/" + x for x in parsed_args.extra]))

        if properties:
            baremetal_client.volume_target.update(
                parsed_args.volume_target, properties)
        else:
            self.log.warning("Please specify what to set.")


class UnsetBaremetalVolumeTarget(command.Command):
    """Unset baremetal volume target properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".UnsetBaremetalVolumeTarget")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'volume_target',
            metavar='<volume target>',
            help=_("UUID of the volume target."))
        parser.add_argument(
            '--extra',
            dest='extra',
            metavar="<key>",
            action='append',
            help=_('Extra to unset (repeat option to unset multiple extras)'))
        parser.add_argument(
            "--property",
            dest='properties',
            metavar="<key>",
            action='append',
            help='Property to unset on this baremetal volume target '
                 '(repeat option to unset multiple properties).',
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch('remove',
                              ['extra/' + x for x in parsed_args.extra]))
        if parsed_args.properties:
            properties.extend(utils.args_array_to_patch(
                'remove', ['properties/' + x for x in parsed_args.properties]))

        if properties:
            baremetal_client.volume_target.update(
                parsed_args.volume_target, properties)
        else:
            self.log.warning("Please specify what to unset.")



# -*- coding: utf-8 -*-
#
# Copyright © 2013 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from typing import Any

from ironicclient.common import base
from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc


class Port(base.Resource):
    def __repr__(self) -> str:
        return "<Port %s>" % self._info


class PortManager(base.CreateManager[Port]):
    resource_class: type[Port] = Port
    _creation_attributes: list[str] = [
        'address', 'extra', 'local_link_connection',
        'node_uuid', 'physical_network', 'portgroup_uuid',
        'pxe_enabled', 'uuid', 'is_smartnic', 'name',
        'description', 'vendor', 'category',
    ]
    _resource_name: str = 'ports'

    def get(
        self,
        port_id: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Port | None:
        return self._get(
            resource_id=port_id,
            fields=fields,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def get_by_address(
        self,
        address: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Port:
        path = '?address=%s' % address
        if fields is not None:
            path += '&fields=' + ','.join(fields)
        else:
            path = 'detail' + path

        ports = self._list(
            self._path(path), 'ports',
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )
        # get all the details of the port assuming that filtering by
        # address returns a collection of one port if successful.
        if len(ports) == 1:
            return ports[0]
        else:
            raise exc.NotFound()

    def delete(
        self,
        port_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        return self._delete(
            resource_id=port_id,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def update(
        self,
        port_id: str,
        patch: list[dict[str, Any]],
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Port:
        return self._update(
            resource_id=port_id,
            patch=patch,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def list(
        self,
        address: str | None = None,
        limit: int | None = None,
        marker: str | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        detail: bool = False,
        fields: list[str] | None = None,
        node: str | None = None,
        portgroup: str | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Port]:
        """Retrieve a list of ports.

        :param address: Optional, MAC address of a port, to get
                       the port which has this MAC address
        :param marker: Optional, the UUID of a port, eg the last
                       port from a previous result set. Return
                       the next result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of ports to return.
            2) limit == 0, return the entire list of ports.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about ports.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param node: Optional, name or UUID of a node. Used to get
                     ports of this node.

        :param portgroup: Optional, name or UUID of a portgroup. Used to get
                          ports of this portgroup.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of ports.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(
                _("Can't fetch a subset of fields "
                  "with 'detail' set"))

        filters = utils.common_filters(
            marker, limit, sort_key, sort_dir, fields)
        if address is not None:
            filters.append('address=%s' % address)
        if node is not None:
            filters.append('node=%s' % node)
        if portgroup is not None:
            filters.append('portgroup=%s' % portgroup)

        path = ''
        if detail:
            path += 'detail'
        if filters:
            path += '?' + '&'.join(filters)
        if limit is None:
            return self._list(
                self._path(path),
                "ports",
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path),
                "ports",
                limit=limit,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import os
from unittest import mock

import dogpile.cache

from ironicclient.common import filecache
from ironicclient.tests.unit import utils


class FileCacheTest(utils.BaseTestCase):

    def test__build_key_ok(self) -> None:
        result = filecache._build_key('localhost', '5000')
        self.assertEqual('localhost:5000', result)

    def test__build_key_none(self) -> None:
        result = filecache._build_key(None, None)
        self.assertEqual('None:None', result)

    @mock.patch.object(filecache, 'CACHE', None)
    @mock.patch.object(os.environ, 'get', autospec=True)
    @mock.patch.object(os.path, 'exists', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    @mock.patch.object(dogpile.cache, 'make_region', autospec=True)
    def test__get_cache_mkdir(self, mock_makeregion: mock.MagicMock,
                              mock_makedirs: mock.MagicMock,
                              mock_exists: mock.MagicMock,
                              mock_get: mock.MagicMock) -> None:
        cache_val = 6
        # If not present in the env, get will return the defaulted value
        mock_get.return_value = filecache.DEFAULT_EXPIRY
        mock_exists.return_value = False
        cache_region = mock.Mock(spec=dogpile.cache.region.CacheRegion)
        cache_region.configure.return_value = cache_val
        mock_makeregion.return_value = cache_region
        self.assertEqual(cache_val, filecache._get_cache())
        self.assertEqual(cache_val, filecache.CACHE)
        mock_exists.assert_called_once_with(filecache.CACHE_DIR)
        mock_makedirs.assert_called_once_with(filecache.CACHE_DIR,
                                              exist_ok=True)
        mock_get.assert_called_once_with(filecache.CACHE_EXPIRY_ENV_VAR,
                                         mock.ANY)
        cache_region.configure.assert_called_once_with(
            mock.ANY,
            arguments=mock.ANY,
            expiration_time=filecache.DEFAULT_EXPIRY)

    @mock.patch.object(filecache, 'CACHE', None)
    @mock.patch.object(os.environ, 'get', autospec=True)
    @mock.patch.object(os.path, 'exists', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    @mock.patch.object(dogpile.cache, 'make_region', autospec=True)
    def test__get_cache_expiry_set(self, mock_makeregion: mock.MagicMock,
                                   mock_makedirs: mock.MagicMock,
                                   mock_exists: mock.MagicMock,
                                   mock_get: mock.MagicMock) -> None:
        cache_val = 5643
        cache_expiry = '78'
        mock_get.return_value = cache_expiry
        mock_exists.return_value = False
        cache_region = mock.Mock(spec=dogpile.cache.region.CacheRegion)
        cache_region.configure.return_value = cache_val
        mock_makeregion.return_value = cache_region
        self.assertEqual(cache_val, filecache._get_cache())
        self.assertEqual(cache_val, filecache.CACHE)
        mock_get.assert_called_once_with(filecache.CACHE_EXPIRY_ENV_VAR,
                                         mock.ANY)
        cache_region.configure.assert_called_once_with(
            mock.ANY,
            arguments=mock.ANY,
            expiration_time=int(cache_expiry))

    @mock.patch.object(filecache, 'CACHE', None)
    @mock.patch.object(filecache.LOG, 'warning', autospec=True)
    @mock.patch.object(os.environ, 'get', autospec=True)
    @mock.patch.object(os.path, 'exists', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    @mock.patch.object(dogpile.cache, 'make_region', autospec=True)
    def test__get_cache_expiry_set_invalid(
        self,
        mock_makeregion: mock.MagicMock,
        mock_makedirs: mock.MagicMock,
        mock_exists: mock.MagicMock,
        mock_get: mock.MagicMock,
        mock_log: mock.MagicMock,
    ) -> None:
        cache_val = 5643
        cache_expiry = 'Rollenhagen'
        mock_get.return_value = cache_expiry
        mock_exists.return_value = False
        cache_region = mock.Mock(spec=dogpile.cache.region.CacheRegion)
        cache_region.configure.return_value = cache_val
        mock_makeregion.return_value = cache_region
        self.assertEqual(cache_val, filecache._get_cache())
        self.assertEqual(cache_val, filecache.CACHE)
        mock_get.assert_called_once_with(filecache.CACHE_EXPIRY_ENV_VAR,
                                         mock.ANY)
        cache_region.configure.assert_called_once_with(
            mock.ANY,
            arguments=mock.ANY,
            expiration_time=filecache.DEFAULT_EXPIRY)
        log_dict = {'curr_val': cache_expiry,
                    'default': filecache.DEFAULT_EXPIRY,
                    'env_var': filecache.CACHE_EXPIRY_ENV_VAR}
        mock_log.assert_called_once_with(mock.ANY, log_dict)

    @mock.patch.object(filecache, 'CACHE', 5552368)
    @mock.patch.object(os.path, 'exists', autospec=True)
    @mock.patch.object(os, 'makedirs', autospec=True)
    def test__get_cache_dir_already_exists(self,
                                           mock_makedirs: mock.MagicMock,
                                           mock_exists: mock.MagicMock,
                                           ) -> None:
        mock_exists.return_value = True
        self.assertEqual(5552368, filecache._get_cache())
        self.assertEqual(5552368, filecache.CACHE)
        self.assertEqual(0, mock_exists.call_count)
        self.assertEqual(0, mock_makedirs.call_count)

    @mock.patch.object(dogpile.cache.region, 'CacheRegion', autospec=True)
    @mock.patch.object(filecache, '_get_cache', autospec=True)
    def test_save_data_ok(self, mock_get_cache: mock.MagicMock,
                          mock_cache: mock.MagicMock) -> None:
        mock_get_cache.return_value = mock_cache
        host = 'fred'
        port = '1234'
        hostport = '%s:%s' % (host, port)
        data = 'some random data'
        filecache.save_data(host, port, data)
        mock_cache.set.assert_called_once_with(hostport, data)

    @mock.patch.object(dogpile.cache.region, 'CacheRegion', autospec=True)
    @mock.patch.object(filecache, '_get_cache', autospec=True)
    def test_save_data_cache_error_swallowed(self,
                                             mock_get_cache: mock.MagicMock,
                                             mock_cache: mock.MagicMock,
                                             ) -> None:
        # A dbm error (e.g. errno 11 under concurrency) must not propagate.
        mock_cache.set.side_effect = OSError(
            11, 'Resource temporarily unavailable')
        mock_get_cache.return_value = mock_cache
        filecache.save_data('fred', '1234', 'some random data')
        mock_cache.set.assert_called_once_with('fred:1234', 'some random data')

    @mock.patch.object(os.path, 'isfile', autospec=True)
    @mock.patch.object(dogpile.cache.region, 'CacheRegion', autospec=True)
    @mock.patch.object(filecache, '_get_cache', autospec=True)
    def test_retrieve_data_cache_error_swallowed(
        self,
        mock_get_cache: mock.MagicMock,
        mock_cache: mock.MagicMock,
        mock_isfile: mock.MagicMock,
    ) -> None:
        # A dbm error while reading is treated as a cache miss.
        mock_isfile.return_value = True
        mock_cache.get.side_effect = OSError(
            11, 'Resource temporarily unavailable')
        mock_get_cache.return_value = mock_cache
        self.assertIsNone(filecache.retrieve_data('fred', '1234'))

    @mock.patch.object(os.path, 'isfile', autospec=True)
    @mock.patch.object(dogpile.cache.region, 'CacheRegion', autospec=True)
    @mock.patch.object(filecache, '_get_cache', autospec=True)
    def test_retrieve_data_ok(self, mock_get_cache: mock.MagicMock,
                              mock_cache: mock.MagicMock,
                              mock_isfile: mock.MagicMock) -> None:
        s = 'spam'
        mock_isfile.return_value = True
        mock_cache.get.return_value = s
        mock_get_cache.return_value = mock_cache
        host = 'fred'
        port = '1234'
        hostport = '%s:%s' % (host, port)
        result = filecache.retrieve_data(host, port)
        mock_cache.get.assert_called_once_with(hostport, expiration_time=None)
        self.assertEqual(s, result)

    @mock.patch.object(os.path, 'isfile', autospec=True)
    @mock.patch.object(dogpile.cache.region, 'CacheRegion', autospec=True)
    @mock.patch.object(filecache, '_get_cache', autospec=True)
    def test_retrieve_data_ok_with_expiry(self,
                                          mock_get_cache: mock.MagicMock,
                                          mock_cache: mock.MagicMock,
                                          mock_isfile: mock.MagicMock,
                                          ) -> None:
        s = 'spam'
        mock_isfile.return_value = True
        mock_cache.get.return_value = s
        mock_get_cache.return_value = mock_cache
        host = 'fred'
        port = '1234'
        expiry = '987'
        hostport = '%s:%s' % (host, port)
        result = filecache.retrieve_data(host, port, expiry)
        mock_cache.get.assert_called_once_with(hostport,
                                               expiration_time=expiry)
        self.assertEqual(s, result)

    @mock.patch.object(os.path, 'isfile', autospec=True)
    @mock.patch.object(dogpile.cache.region, 'CacheRegion', autospec=True)
    @mock.patch.object(filecache, '_get_cache', autospec=True)
    def test_retrieve_data_not_found(self,
                                     mock_get_cache: mock.MagicMock,
                                     mock_cache: mock.MagicMock,
                                     mock_isfile: mock.MagicMock,
                                     ) -> None:
        mock_isfile.return_value = True
        mock_cache.get.return_value = dogpile.cache.api.NO_VALUE
        mock_get_cache.return_value = mock_cache
        host = 'fred'
        port = '1234'
        hostport = '%s:%s' % (host, port)
        result = filecache.retrieve_data(host, port)
        mock_cache.get.assert_called_once_with(hostport, expiration_time=None)
        self.assertIsNone(result)

    @mock.patch.object(os.path, 'isfile', autospec=True)
    def test_retrieve_data_no_cache_file(self,
                                         mock_isfile: mock.MagicMock,
                                         ) -> None:
        mock_isfile.return_value = False
        self.assertIsNone(filecache.retrieve_data(host='spam', port='eggs'))



#    Copyright (c) 2015 Intel Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import testtools

from ironicclient.v1 import resource_fields


class ResourceTest(testtools.TestCase):
    def setUp(self) -> None:
        super(ResourceTest, self).setUp()
        self._saved_ids = resource_fields.Resource.FIELDS
        resource_fields.Resource.FIELDS = {
            'item1': 'ITEM1',
            '2nd_item': 'A second item',
            'item_3': 'Third item',
        }

    def tearDown(self) -> None:
        super(ResourceTest, self).tearDown()
        resource_fields.Resource.FIELDS = self._saved_ids

    def test_fields_single_value(self) -> None:
        # Make sure single value is what we expect
        foo = resource_fields.Resource(['item1'])
        self.assertEqual(('item1',), foo.fields)
        self.assertEqual(('ITEM1',), foo.labels)
        self.assertEqual(('item1',), foo.sort_fields)
        self.assertEqual(('ITEM1',), foo.sort_labels)

    def test_fields_multiple_value_order(self) -> None:
        # Make sure order is maintained
        foo = resource_fields.Resource(['2nd_item', 'item1'])
        self.assertEqual(('2nd_item', 'item1'), foo.fields)
        self.assertEqual(('A second item', 'ITEM1'), foo.labels)
        self.assertEqual(('2nd_item', 'item1'), foo.sort_fields)
        self.assertEqual(('A second item', 'ITEM1'), foo.sort_labels)

    def test_sort_excluded(self) -> None:
        # Test excluding of fields for sort purposes
        foo = resource_fields.Resource(['item_3', 'item1', '2nd_item'],
                                       sort_excluded=['item1'])
        self.assertEqual(('item_3', '2nd_item'), foo.sort_fields)
        self.assertEqual(('Third item', 'A second item'), foo.sort_labels)

    def test_sort_excluded_unknown(self) -> None:
        # Test sort_excluded value not in the field_ids
        self.assertRaises(
            ValueError,
            resource_fields.Resource,
            ['item_3', 'item1', '2nd_item'],
            sort_excluded=['item1', 'foo'])

    def test_override_labels(self) -> None:
        # Test overriding labels
        foo = resource_fields.Resource(['item_3', 'item1', '2nd_item'],
                                       override_labels={'item1': 'One'})
        self.assertEqual(('Third item', 'One', 'A second item'), foo.labels)
        self.assertEqual(('Third item', 'One', 'A second item'),
                         foo.sort_labels)

    def test_override_labels_unknown(self) -> None:
        # Test overriding labels with key not in field_ids
        self.assertRaises(
            ValueError,
            resource_fields.Resource,
            ['item_3', 'item1', '2nd_item'],
            override_labels={'foo': 'One'})

    def test_sort_excluded_override_labels(self) -> None:
        # Test excluding of fields for sort purposes and overriding labels
        foo = resource_fields.Resource(['item_3', 'item1', '2nd_item'],
                                       sort_excluded=['item1'],
                                       override_labels={'item_3': 'Three'})
        self.assertEqual(('Three', 'ITEM1', 'A second item'), foo.labels)
        self.assertEqual(('item_3', '2nd_item'), foo.sort_fields)
        self.assertEqual(('Three', 'A second item'), foo.sort_labels)

    def test_unknown_field_id(self) -> None:
        self.assertRaises(
            KeyError,
            resource_fields.Resource,
            ['item1', 'unknown_id'])



#
#   Copyright 2016 Mirantis, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
import itertools
import logging
from typing import Any, cast

from osc_lib import utils as oscutils

from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc
from ironicclient.osc import command
from ironicclient.v1 import resource_fields as res_fields


class CreateBaremetalPortGroup(command.ShowOne):
    """Create a new baremetal port group."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".CreateBaremetalPortGroup")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--node',
            dest='node_uuid',
            metavar='<uuid>',
            required=True,
            help=_('UUID of the node that this port group belongs to.'))
        parser.add_argument(
            '--address',
            metavar='<mac-address>',
            help=_('MAC address for this port group.'))
        parser.add_argument(
            '--name',
            dest='name',
            help=_('Name of the port group.'))
        parser.add_argument(
            '--uuid',
            dest='uuid',
            help=_('UUID of the port group.'))
        parser.add_argument(
            '--extra',
            metavar="<key=value>",
            action='append',
            help=_("Record arbitrary key/value metadata. "
                   "Can be specified multiple times."))
        parser.add_argument(
            '--mode',
            help=_('Mode of the port group. For possible values, refer to '
                   'https://www.kernel.org/doc/Documentation/networking'
                   '/bonding.txt.'))
        parser.add_argument(
            '--property',
            dest='properties',
            metavar="<key=value>",
            action='append',
            help=_("Key/value property related to this port group's "
                   "configuration. Can be specified multiple times."))
        standalone_ports_group = parser.add_mutually_exclusive_group()
        standalone_ports_group.add_argument(
            '--support-standalone-ports',
            action='store_true',
            help=_("Ports that are members of this port group "
                   "can be used as stand-alone ports. (default)"))
        standalone_ports_group.add_argument(
            '--unsupport-standalone-ports',
            action='store_true',
            help=_("Ports that are members of this port group "
                   "cannot be used as stand-alone ports."))
        parser.add_argument(
            '--physical-network',
            dest='physical_network',
            metavar='<physical network>',
            help=_("Name of the physical network to which the ports in "
                   "this port group are connected."))
        parser.add_argument(
            '--category',
            dest='category',
            metavar='<port category>',
            help=_("An optional category for the port group."))

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        field_list: list[str] = [
            'node_uuid', 'address', 'name', 'uuid', 'extra', 'mode',
            'properties', 'physical_network', 'category']
        fields = dict((k, v) for (k, v) in vars(parsed_args).items()
                      if k in field_list and v is not None)
        if parsed_args.support_standalone_ports:
            fields['standalone_ports_supported'] = True
        if parsed_args.unsupport_standalone_ports:
            fields['standalone_ports_supported'] = False

        fields = utils.args_array_to_dict(fields, 'extra')
        fields = utils.args_array_to_dict(fields, 'properties')
        portgroup = baremetal_client.portgroup.create(**fields)

        data = dict([(f, getattr(portgroup, f, '')) for f in
                     res_fields.PORTGROUP_DETAILED_RESOURCE.fields])

        return self.dict2columns(data)


class ShowBaremetalPortGroup(command.ShowOne):
    """Show baremetal port group details."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ShowBaremetalPortGroup")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "portgroup",
            metavar="<id>",
            help=_("UUID or name of the port group "
                   "(or MAC address if --address is specified)."))
        parser.add_argument(
            '--address',
            dest='address',
            action='store_true',
            default=False,
            help=_('<id> is the MAC address (instead of UUID or name) '
                   'of the port group.'))
        parser.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            choices=res_fields.PORTGROUP_DETAILED_RESOURCE.fields,
            default=[],
            help=_("One or more port group fields. Only these fields will be "
                   "fetched from the server."))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal
        fields: list[str] | None
        fields = (
            list(itertools.chain.from_iterable(parsed_args.fields))
            or None)

        if parsed_args.address:
            portgroup = baremetal_client.portgroup.get_by_address(
                parsed_args.portgroup, fields=fields)._info
        else:
            portgroup = baremetal_client.portgroup.get(
                parsed_args.portgroup, fields=fields)._info

        portgroup.pop("links", None)
        portgroup.pop("ports", None)
        return cast(
            tuple[tuple[str, ...], tuple[Any, ...]],
            tuple(zip(*sorted(portgroup.items()))),
        )


class ListBaremetalPortGroup(command.Lister):
    """List baremetal port groups."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalPortGroup")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of port groups to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Baremetal API Service.'))
        parser.add_argument(
            '--marker',
            metavar='<port group>',
            help=_('Port group UUID (for example, of the last port group in '
                   'the list from a previous request). Returns the list of '
                   'port groups after this UUID.'))
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            help=_('Sort output by specified port group fields and directions '
                   '(asc or desc) (default: asc). Multiple fields and '
                   'directions can be specified, separated by comma.'))
        parser.add_argument(
            '--address',
            metavar='<mac-address>',
            help=_("Only show information for the port group with this MAC "
                   "address."))
        parser.add_argument(
            '--node',
            dest='node',
            metavar='<node>',
            help=_("Only list port groups of this node (name or UUID)."))
        parser.add_argument(
            '--shards',
            nargs='+',
            metavar='<shard>',
            help=_("Only list port groups of nodes in these shards."))

        display_group = parser.add_mutually_exclusive_group(required=False)
        display_group.add_argument(
            '--long',
            default=False,
            dest='detail',
            help=_("Show detailed information about the port groups."),
            action='store_true')
        display_group.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.PORTGROUP_DETAILED_RESOURCE.fields,
            help=_("One or more port group fields. Only these fields will be "
                   "fetched from the server. Can not be used when '--long' is "
                   "specified."))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        client = manager.baremetal

        columns = res_fields.PORTGROUP_RESOURCE.fields

        params: dict[str, object] = {}
        if parsed_args.limit is not None and parsed_args.limit < 0:
            raise exc.CommandError(
                _('Expected non-negative --limit, got %s') %
                parsed_args.limit)
        params['limit'] = parsed_args.limit
        params['marker'] = parsed_args.marker
        if parsed_args.address is not None:
            params['address'] = parsed_args.address
        if parsed_args.node is not None:
            params['node'] = parsed_args.node
        if parsed_args.shards:
            params['shards'] = parsed_args.shards

        if parsed_args.detail:
            params['detail'] = parsed_args.detail
            columns = res_fields.PORTGROUP_DETAILED_RESOURCE.fields
        elif parsed_args.fields:
            params['detail'] = False
            fields = itertools.chain.from_iterable(parsed_args.fields)
            resource = res_fields.Resource(list(fields))
            columns = resource.fields
            params['fields'] = columns

        self.log.debug("params(%s)", params)
        data = client.portgroup.list(**params)

        data = oscutils.sort_items(data, parsed_args.sort)

        return (columns,
                (oscutils.get_item_properties(s, columns, formatters={
                    'Properties': utils.HashColumn},) for s in data))


class DeleteBaremetalPortGroup(command.Command):
    """Unregister baremetal port group(s)."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".DeleteBaremetalPortGroup")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "portgroups",
            metavar="<port group>",
            nargs="+",
            help=_("Port group(s) to delete (name or UUID)."))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        failures: list[str] = []
        for portgroup in parsed_args.portgroups:
            try:
                baremetal_client.portgroup.delete(portgroup)
                print(_('Deleted port group %s') % portgroup)
            except exc.ClientException as e:
                failures.append(_("Failed to delete port group %(portgroup)s: "
                                  " %(error)s")
                                % {'portgroup': portgroup, 'error': e})

        if failures:
            raise exc.ClientException("\n".join(failures))


class SetBaremetalPortGroup(command.Command):
    """Set baremetal port group properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".SetBaremetalPortGroup")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'portgroup',
            metavar='<port group>',
            help=_("Name or UUID of the port group."),
        )
        parser.add_argument(
            '--node',
            dest='node_uuid',
            metavar='<uuid>',
            help=_('Update UUID of the node that this port group belongs to.')
        )
        parser.add_argument(
            "--address",
            metavar="<mac-address>",
            help=_("MAC address for this port group."),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            help=_("Name of the port group."),
        )
        parser.add_argument(
            "--extra",
            metavar="<key=value>",
            action='append',
            help=_('Extra to set on this baremetal port group '
                   '(repeat option to set multiple extras).'),
        )
        parser.add_argument(
            '--mode',
            help=_('Mode of the port group. For possible values, refer to '
                   'https://www.kernel.org/doc/Documentation/networking'
                   '/bonding.txt.'))
        parser.add_argument(
            '--property',
            dest='properties',
            metavar="<key=value>",
            action='append',
            help=_("Key/value property related to this port group's "
                   "configuration (repeat option to set multiple "
                   "properties)."))
        standalone_ports_group = parser.add_mutually_exclusive_group()
        standalone_ports_group.add_argument(
            '--support-standalone-ports',
            action='store_true',
            default=None,
            help=_("Ports that are members of this port group "
                   "can be used as stand-alone ports.")
        )
        standalone_ports_group.add_argument(
            '--unsupport-standalone-ports',
            action='store_true',
            help=_("Ports that are members of this port group "
                   "cannot be used as stand-alone ports.")
        )
        parser.add_argument(
            '--physical-network',
            metavar='<physical network>',
            dest='physical_network',
            help=_("Set a physical network for the ports in this port group"))

        parser.add_argument(
            '--category',
            metavar='<category>',
            dest='category',
            help=_("Set a category for this port group"))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        if parsed_args.node_uuid:
            properties.extend(utils.args_array_to_patch(
                'add', ["node_uuid=%s" % parsed_args.node_uuid]))
        if parsed_args.address:
            properties.extend(utils.args_array_to_patch(
                'add', ["address=%s" % parsed_args.address]))
        if parsed_args.name:
            name = ["name=%s" % parsed_args.name]
            properties.extend(utils.args_array_to_patch(
                'add', name))
        if parsed_args.physical_network:
            physical_network = [
                f"physical_network={parsed_args.physical_network}"
            ]
            properties.extend(utils.args_array_to_patch(
                'add', physical_network))
        if parsed_args.category:
            category = [f"category={parsed_args.category}"]
            properties.extend(utils.args_array_to_patch(
                'add', category))
        if parsed_args.support_standalone_ports:
            properties.extend(utils.args_array_to_patch(
                'add', ["standalone_ports_supported=True"]))
        if parsed_args.unsupport_standalone_ports:
            properties.extend(utils.args_array_to_patch(
                'add', ["standalone_ports_supported=False"]))
        if parsed_args.mode:
            properties.extend(utils.args_array_to_patch(
                'add', ["mode=\"%s\"" % parsed_args.mode]))

        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch(
                'add', ['extra/' + x for x in parsed_args.extra]))
        if parsed_args.properties:
            properties.extend(utils.args_array_to_patch(
                'add', ['properties/' + x for x in parsed_args.properties]))

        if properties:
            baremetal_client.portgroup.update(parsed_args.portgroup,
                                              properties)
        else:
            self.log.warning("Please specify what to set.")


class UnsetBaremetalPortGroup(command.Command):
    """Unset baremetal port group properties."""
    log: logging.Logger = logging.getLogger(
        __name__ + ".UnsetBaremetalPortGroup")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'portgroup',
            metavar='<port group>',
            help=_("Name or UUID of the port group.")
        )
        parser.add_argument(
            "--name",
            action='store_true',
            help=_("Unset the name of the port group."),
        )
        parser.add_argument(
            "--address",
            action='store_true',
            help=_("Unset the address of the port group."),
        )
        parser.add_argument(
            "--extra",
            metavar="<key>",
            action='append',
            help=_('Extra to unset on this baremetal port group '
                   '(repeat option to unset multiple extras).'),
        )
        parser.add_argument(
            "--property",
            dest='properties',
            metavar="<key>",
            action='append',
            help=_('Property to unset on this baremetal port group '
                   '(repeat option to unset multiple properties).'),
        )
        parser.add_argument(
            '--physical-network',
            action='store_true',
            dest='physical_network',
            help=_("Unset the physical network on this baremetal port group."))

        parser.add_argument(
            '--category',
            dest='category',
            action='store_true',
            help=_("Unset the category for this port group."))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        if parsed_args.name:
            properties.extend(utils.args_array_to_patch('remove',
                              ['name']))
        if parsed_args.address:
            properties.extend(utils.args_array_to_patch('remove',
                              ['address']))
        if parsed_args.physical_network:
            properties.extend(utils.args_array_to_patch('remove',
                              ['physical_network']))
        if parsed_args.category:
            properties.extend(utils.args_array_to_patch('remove',
                              ['category']))
        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch('remove',
                              ['extra/' + x for x in parsed_args.extra]))
        if parsed_args.properties:
            properties.extend(utils.args_array_to_patch(
                'remove', ['properties/' + x for x in parsed_args.properties]))

        if properties:
            baremetal_client.portgroup.update(parsed_args.portgroup,
                                              properties)
        else:
            self.log.warning("Please specify what to unset.")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import requests

from ironicclient.common.apiclient import exceptions
from ironicclient.common.apiclient.exceptions import *  # noqa


# NOTE(akurilin): This alias is left here since v.0.1.3 to support backwards
# compatibility.
InvalidEndpoint = exceptions.EndpointException
CommunicationError = exceptions.ConnectionRefused
HTTPBadRequest = exceptions.BadRequest
HTTPInternalServerError = exceptions.InternalServerError
HTTPNotFound = exceptions.NotFound
HTTPServiceUnavailable = exceptions.ServiceUnavailable


class AmbiguousAuthSystem(exceptions.ClientException):
    """Could not obtain token and endpoint using provided credentials."""
    pass


# Alias for backwards compatibility
AmbigiousAuthSystem = AmbiguousAuthSystem


class InvalidAttribute(exceptions.ClientException):
    pass


class StateTransitionFailed(exceptions.ClientException):
    """Failed to reach a requested provision state."""


class StateTransitionTimeout(exceptions.ClientException):
    """Timed out while waiting for a requested provision state."""


def from_response(  # type: ignore[no-redef]
    response: requests.Response,
    message: str | None = None,
    traceback: str | None = None,
    method: str | None = None,
    url: str | None = None,
) -> exceptions.HttpError:
    """Return an HttpError instance based on response from httplib/requests."""

    error_body: dict[str, str] = {}
    if message:
        error_body['message'] = message
    if traceback:
        error_body['details'] = traceback

    if hasattr(response, 'status') and not hasattr(response, 'status_code'):
        # NOTE(akurilin): These modifications around response object give
        # ability to get all necessary information in method `from_response`
        # from common code, which expecting response object from `requests`
        # library instead of object from `httplib/httplib2` library.
        response.status_code = response.status
        response.headers = {  # type: ignore[assignment]
            'Content-Type':
            response.getheader(  # type: ignore[attr-defined]
                'content-type', "")}

    if hasattr(response, 'status_code'):
        # NOTE(jiangfei): These modifications allow SessionClient
        # to handle faultstring.
        response.json = (  # type: ignore[method-assign]
            lambda: {  # type: ignore[misc,assignment]
                'error': error_body})

    return exceptions.from_response(
        response, method=method or "", url=url or ""
    )



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import copy
import io
import os
from collections.abc import ItemsView, Iterator
from typing import Any
from unittest import mock

import fixtures
from oslo_utils import strutils
import requests
import testtools


class BaseTestCase(testtools.TestCase):
    def setUp(self) -> None:
        super(BaseTestCase, self).setUp()
        self.useFixture(fixtures.FakeLogger())

        # If enabled, stdout and/or stderr is captured and will appear in
        # test results if that test fails.
        if strutils.bool_from_string(os.environ.get('OS_STDOUT_CAPTURE')):
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if strutils.bool_from_string(os.environ.get('OS_STDERR_CAPTURE')):
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))


class FakeAPI(object):
    def __init__(
        self,
        responses: dict[
            str, dict[str, tuple[dict[str, str], Any]]
        ],
        path_prefix: str | None = None,
    ) -> None:
        self.responses = responses
        self.calls: list[tuple[Any, ...]] = []
        self.path_prefix = path_prefix or ''
        self.endpoint_trimmed = (
            'http://127.0.0.1:6385' + self.path_prefix)

    def _request(
        self,
        method: str,
        url: str,
        headers: dict[str, str] | None = None,
        body: Any = None,
        params: dict[str, Any] | None = None,
    ) -> tuple[dict[str, str], Any]:
        # url should always just be a path here, e.g. /v1/nodes
        url = self.path_prefix + url

        if params:
            self.calls.append((method, url, headers or {}, body, params))
        else:
            self.calls.append((method, url, headers or {}, body))
        return self.responses[url][method]

    def raw_request(
        self, *args: Any, **kwargs: Any,
    ) -> tuple[FakeResponse, Iterator[str]]:
        response = self._request(*args, **kwargs)
        body_iter = iter(io.StringIO(response[1]))
        return FakeResponse(response[0]), body_iter

    def json_request(
        self, *args: Any, **kwargs: Any,
    ) -> tuple[FakeResponse, Any]:
        response = self._request(*args, **kwargs)
        return FakeResponse(response[0]), response[1]


class FakeConnection(object):
    def __init__(self, response: Any = None,
                 path_prefix: str | None = None) -> None:
        self._response = response
        self._last_request: tuple[str, str, dict[str, Any]] | None = None

    def request(self, method: str, conn_url: str,
                **kwargs: Any) -> None:
        self._last_request = (method, conn_url, kwargs)

    def setresponse(self, response: Any) -> None:
        self._response = response

    def getresponse(self) -> Any:
        return self._response

    def __repr__(self) -> str:
        return ("FakeConnection(response=%s)" % (self._response))


class FakeResponse(object):
    def __init__(
        self,
        headers: dict[str, str],
        body: io.StringIO | None = None,
        version: str | None = None,
        status: int | None = None,
        reason: str | None = None,
        request_headers: dict[str, str] | None = None,
    ) -> None:
        """Fake object to help testing.

        :param headers: dict representing HTTP response headers
        :param body: file-like object
        """
        self.headers = headers
        self.body = body
        self.raw = mock.Mock()
        self.raw.version = version
        self.status_code = status
        self.reason = reason
        self.request = mock.Mock()
        self.request.headers = request_headers or {}

    def getheaders(self) -> ItemsView[str, str]:
        return copy.deepcopy(self.headers).items()

    def getheader(self, key: str, default: str) -> str:
        return self.headers.get(key, default)

    def read(self, amt: int) -> str:
        return self.body.read(amt)

    def __repr__(self) -> str:
        return ("FakeResponse(%s, body=%s, version=%s, status=%s, "
                "reason=%s, request_headers=%s)" %
                (self.headers, self.body, self.raw.version, self.status_code,
                 self.reason, self.request.headers))


def mockSessionResponse(
    headers: dict[str, str],
    content: str | None = None,
    status_code: int | None = None,
    version: str | None = None,
    request_headers: dict[str, str] | None = None,
) -> mock.Mock:
    raw = mock.Mock()
    raw.version = version
    request = mock.Mock()
    request.headers = request_headers or {}
    response = mock.Mock(spec=requests.Response,
                         headers=headers,
                         content=content,
                         status_code=status_code,
                         raw=raw,
                         reason='',
                         encoding='UTF-8',
                         request=request)
    response.text = content

    return response


def mockSession(
    headers: dict[str, str],
    content: str | None = None,
    status_code: int | None = None,
    version: str | None = None,
) -> mock.Mock:
    session = mock.Mock(spec=requests.Session,
                        verify=False,
                        cert=('test_cert', 'test_key'))
    session.get_endpoint = mock.Mock(return_value='https://test')
    response = mockSessionResponse(headers, content, status_code, version)
    session.request = mock.Mock(return_value=response)

    return session



#
#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
import itertools
import logging
from typing import Any, cast

from osc_lib import utils as oscutils

from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc
from ironicclient.osc import command
from ironicclient.v1 import resource_fields as res_fields


class CreateBaremetalPort(command.ShowOne):
    """Create a new port"""

    log: logging.Logger = logging.getLogger(
        __name__ + ".CreateBaremetalPort")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'address',
            metavar='<address>',
            help=_('MAC address for this port.')
        )
        parser.add_argument(
            '--node',
            dest='node_uuid',
            metavar='<uuid>',
            required=True,
            help=_('UUID of the node that this port belongs to.')
        )
        parser.add_argument(
            '--uuid',
            dest='uuid',
            metavar='<uuid>',
            help=_('UUID of the port.'))
        parser.add_argument(
            '--extra',
            metavar="<key=value>",
            action='append',
            help=_("Record arbitrary key/value metadata. "
                   "Argument can be specified multiple times.")
        )
        parser.add_argument(
            '--local-link-connection',
            metavar="<key=value>",
            action='append',
            help=_("Key/value metadata describing Local link connection "
                   "information. Valid keys are 'switch_info', 'switch_id', "
                   "'port_id' and 'hostname'. The keys 'switch_id' and "
                   "'port_id' are required. In case of a Smart NIC port, "
                   "the required keys are 'port_id' and 'hostname'. "
                   "Argument can be specified multiple times.")
        )
        parser.add_argument(
            '--pxe-enabled',
            metavar='<boolean>',
            help=_('Indicates whether this Port should be used when '
                   'PXE booting this Node.')
        )

        parser.add_argument(
            '--port-group',
            dest='portgroup_uuid',
            metavar='<uuid>',
            help=_("UUID of the port group that this port belongs to."))

        parser.add_argument(
            '--physical-network',
            dest='physical_network',
            metavar='<physical network>',
            help=_("Name of the physical network to which this port is "
                   "connected."))

        parser.add_argument(
            '--is-smartnic',
            dest='is_smartnic',
            action='store_true',
            help=_("Indicates whether this Port is a Smart NIC port"))

        parser.add_argument(
            '--name',
            dest='name',
            metavar='<port name>',
            help=_("Name of the port."))

        parser.add_argument(
            '--description',
            dest='description',
            metavar='<port description>',
            help=_("An optional description for the port."))

        parser.add_argument(
            '--vendor',
            dest='vendor',
            metavar='<port vendor>',
            help=_("An optional vendor for the port."))

        parser.add_argument(
            '--category',
            dest='category',
            metavar='<port category>',
            help=_("An optional category for the port."))

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        field_list: list[str] = [
            'address', 'uuid', 'extra', 'node_uuid', 'pxe_enabled',
            'local_link_connection', 'portgroup_uuid',
            'physical_network', 'name', 'description', 'vendor',
            'category']
        fields = dict((k, v) for (k, v) in vars(parsed_args).items()
                      if k in field_list and v is not None)
        fields = utils.args_array_to_dict(fields, 'extra')
        fields = utils.args_array_to_dict(fields, 'local_link_connection')
        if parsed_args.is_smartnic:
            fields['is_smartnic'] = parsed_args.is_smartnic
        port = baremetal_client.port.create(**fields)

        data = dict([(f, getattr(port, f, '')) for f in
                     res_fields.PORT_DETAILED_RESOURCE.fields])

        return self.dict2columns(data)


class ShowBaremetalPort(command.ShowOne):
    """Show baremetal port details."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ShowBaremetalPort")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "port",
            metavar="<id>",
            help=_("UUID of the port (or MAC address if --address is "
                   "specified).")
        )
        parser.add_argument(
            '--address',
            dest='address',
            action='store_true',
            default=False,
            help=_('<id> is the MAC address (instead of the UUID) of the '
                   'port.')
        )
        parser.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            choices=res_fields.PORT_DETAILED_RESOURCE.fields,
            default=[],
            help=_("One or more port fields. Only these fields will be "
                   "fetched from the server.")
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal
        fields: list[str] | None
        fields = (
            list(itertools.chain.from_iterable(parsed_args.fields))
            or None)

        if parsed_args.address:
            port = baremetal_client.port.get_by_address(
                parsed_args.port, fields=fields)._info
        else:
            port = baremetal_client.port.get(
                parsed_args.port, fields=fields)._info

        port.pop("links", None)
        return cast(
            tuple[tuple[str, ...], tuple[Any, ...]],
            tuple(zip(*sorted(port.items()))),
        )


class UnsetBaremetalPort(command.Command):
    """Unset baremetal port properties."""
    log: logging.Logger = logging.getLogger(
        __name__ + ".UnsetBaremetalPort")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'port',
            metavar='<port>',
            help=_("UUID of the port.")
        )
        parser.add_argument(
            "--extra",
            metavar="<key>",
            action='append',
            help=_('Extra to unset on this baremetal port '
                   '(repeat option to unset multiple extras)')
        )

        parser.add_argument(
            "--local-link-connection",
            metavar="<key>",
            action='append',
            help=_("Remove local link connection on this port "
                   "(repeat option to unset multiple keys)"))

        parser.add_argument(
            '--port-group',
            action='store_true',
            dest='portgroup',
            help=_("Remove port from the port group"))

        parser.add_argument(
            '--physical-network',
            action='store_true',
            dest='physical_network',
            help=_("Unset the physical network on this baremetal port."))

        parser.add_argument(
            '--is-smartnic',
            dest='is_smartnic',
            action='store_true',
            help=_("Set Port as not Smart NIC port"))

        parser.add_argument(
            '--name',
            dest='name',
            action='store_true',
            help=_("Unset the name for this port"))

        parser.add_argument(
            '--description',
            dest='description',
            action='store_true',
            help=_("Unset the description for this port."))

        parser.add_argument(
            '--vendor',
            dest='vendor',
            action='store_true',
            help=_("Unset the vendor for this port."))

        parser.add_argument(
            '--category',
            dest='category',
            action='store_true',
            help=_("Unset the category for this port."))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal
        properties: list[dict[str, Any]] = []
        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch(
                'remove',
                ['extra/' + x for x in parsed_args.extra]))
        if parsed_args.local_link_connection:
            properties.extend(utils.args_array_to_patch(
                'remove', ['local_link_connection/' + x for x in
                           parsed_args.local_link_connection]))
        if parsed_args.portgroup:
            properties.extend(utils.args_array_to_patch('remove',
                              ['portgroup_uuid']))
        if parsed_args.physical_network:
            properties.extend(utils.args_array_to_patch('remove',
                              ['physical_network']))
        if parsed_args.is_smartnic:
            properties.extend(utils.args_array_to_patch(
                'add', ["is_smartnic=False"]))
        if parsed_args.name:
            properties.extend(utils.args_array_to_patch('remove',
                              ['name']))
        if parsed_args.description:
            properties.extend(utils.args_array_to_patch('remove',
                              ['description']))
        if parsed_args.vendor:
            properties.extend(utils.args_array_to_patch('remove',
                              ['vendor']))
        if parsed_args.category:
            properties.extend(utils.args_array_to_patch('remove',
                              ['category']))

        if properties:
            baremetal_client.port.update(parsed_args.port, properties)
        else:
            self.log.warning("Please specify what to unset.")


class SetBaremetalPort(command.Command):
    """Set baremetal port properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".SetBaremetalPort")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'port',
            metavar='<port>',
            help=_("UUID of the port")
        )
        parser.add_argument(
            '--node',
            dest='node_uuid',
            metavar='<uuid>',
            help=_('Set UUID of the node that this port belongs to')
        )
        parser.add_argument(
            "--address",
            metavar="<address>",
            dest='address',
            help=_("Set MAC address for this port")
        )
        parser.add_argument(
            "--extra",
            metavar="<key=value>",
            action='append',
            help=_('Extra to set on this baremetal port '
                   '(repeat option to set multiple extras)')
        )
        parser.add_argument(
            "--port-group",
            metavar="<uuid>",
            dest='portgroup_uuid',
            help=_('Set UUID of the port group that this port belongs to.'))
        parser.add_argument(
            "--local-link-connection",
            metavar="<key=value>",
            action='append',
            help=_("Key/value metadata describing local link connection "
                   "information. Valid keys are 'switch_info', 'switch_id', "
                   "'port_id' and 'hostname'. The keys 'switch_id' and "
                   "'port_id' are required. In case of a Smart NIC port, "
                   "the required keys are 'port_id' and 'hostname'. "
                   "Argument can be specified multiple times.")
        )
        pxe_enabled_group = parser.add_mutually_exclusive_group(required=False)
        pxe_enabled_group.add_argument(
            "--pxe-enabled",
            dest='pxe_enabled',
            default=None,
            action='store_true',
            help=_("Indicates that this port should be used when "
                   "PXE booting this node (default)")
        )
        pxe_enabled_group.add_argument(
            "--pxe-disabled",
            dest='pxe_enabled',
            default=None,
            action='store_false',
            help=_("Indicates that this port should not be used when "
                   "PXE booting this node")
        )
        parser.add_argument(
            '--physical-network',
            metavar='<physical network>',
            dest='physical_network',
            help=_("Set the name of the physical network to which this port "
                   "is connected."))

        parser.add_argument(
            '--is-smartnic',
            dest='is_smartnic',
            action='store_true',
            help=_("Set port to be Smart NIC port"))

        parser.add_argument(
            '--name',
            metavar='<name>',
            dest='name',
            help=_("Set name for this port"))

        parser.add_argument(
            '--description',
            metavar='<description>',
            dest='description',
            help=_("Set a description for this port"))

        parser.add_argument(
            '--vendor',
            metavar='<vendor>',
            dest='vendor',
            help=_("Set a vendor for this port"))

        parser.add_argument(
            '--category',
            metavar='<category>',
            dest='category',
            help=_("Set a category for this port"))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        if parsed_args.node_uuid:
            node_uuid = ["node_uuid=%s" % parsed_args.node_uuid]
            properties.extend(utils.args_array_to_patch(
                'add', node_uuid))
        if parsed_args.address:
            address = ["address=%s" % parsed_args.address]
            properties.extend(utils.args_array_to_patch('add', address))
        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch(
                'add', ['extra/' + x for x in parsed_args.extra]))
        if parsed_args.portgroup_uuid:
            portgroup_uuid = ["portgroup_uuid=%s" % parsed_args.portgroup_uuid]
            properties.extend(utils.args_array_to_patch('add', portgroup_uuid))
        if parsed_args.local_link_connection:
            properties.extend(utils.args_array_to_patch(
                'add', ['local_link_connection/' + x for x in
                        parsed_args.local_link_connection]))
        if parsed_args.pxe_enabled is not None:
            properties.extend(utils.args_array_to_patch(
                'add', ['pxe_enabled=%s' % parsed_args.pxe_enabled]))
        if parsed_args.physical_network:
            physical_network = ["physical_network=%s" %
                                parsed_args.physical_network]
            properties.extend(utils.args_array_to_patch('add',
                                                        physical_network))
        if parsed_args.is_smartnic:
            is_smartnic = ["is_smartnic=%s" % parsed_args.is_smartnic]
            properties.extend(utils.args_array_to_patch('add', is_smartnic))
        if parsed_args.name:
            port_name = ["name=%s" % parsed_args.name]
            properties.extend(utils.args_array_to_patch('add', port_name))
        if parsed_args.description:
            port_description = ["description=%s" % parsed_args.description]
            properties.extend(utils.args_array_to_patch('add',
                                                        port_description))
        if parsed_args.vendor:
            port_vendor = ["vendor=%s" % parsed_args.vendor]
            properties.extend(utils.args_array_to_patch('add',
                                                        port_vendor))
        if parsed_args.category:
            port_category = ["category=%s" % parsed_args.category]
            properties.extend(utils.args_array_to_patch('add',
                                                        port_category))

        if properties:
            baremetal_client.port.update(parsed_args.port, properties)
        else:
            self.log.warning("Please specify what to set.")


class DeleteBaremetalPort(command.Command):
    """Delete port(s)."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".DeleteBaremetalPort")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "ports",
            metavar="<port>",
            nargs="+",
            help=_("UUID(s) of the port(s) to delete.")
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        failures: list[str] = []
        for port in parsed_args.ports:
            try:
                baremetal_client.port.delete(port)
                print(_('Deleted port %s') % port)
            except exc.ClientException as e:
                failures.append(_("Failed to delete port %(port)s: %(error)s")
                                % {'port': port, 'error': e})

        if failures:
            raise exc.ClientException("\n".join(failures))


class ListBaremetalPort(command.Lister):
    """List baremetal ports."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalPort")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--address',
            dest='address',
            metavar='<mac-address>',
            help=_("Only show information for the port with this MAC address.")
        )
        parser.add_argument(
            '--node',
            dest='node',
            metavar='<node>',
            help=_("Only list ports of this node (name or UUID).")
        )
        parser.add_argument(
            "--port-group",
            metavar="<port group>",
            dest='portgroup',
            help=_('Only list ports of this port group (name or UUID).'))
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of ports to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Baremetal API Service.')
        )
        parser.add_argument(
            '--marker',
            metavar='<port>',
            help=_('Port UUID (for example, of the last port in the list '
                   'from a previous request). Returns the list of ports '
                   'after this UUID.')
        )
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            help=_('Sort output by specified port fields and directions '
                   '(asc or desc) (default: asc). Multiple fields and '
                   'directions can be specified, separated by comma.')
        )
        display_group = parser.add_mutually_exclusive_group()
        display_group.add_argument(
            '--long',
            dest='detail',
            action='store_true',
            default=False,
            help=_("Show detailed information about ports.")
        )
        display_group.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.PORT_DETAILED_RESOURCE.fields,
            help=_("One or more port fields. Only these fields will be "
                   "fetched from the server. Can not be used when "
                   "'--long' is specified.")
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        client = manager.baremetal

        columns = res_fields.PORT_RESOURCE.fields

        params: dict[str, object] = {}
        if parsed_args.limit is not None and parsed_args.limit < 0:
            raise exc.CommandError(
                _('Expected non-negative --limit, got %s') %
                parsed_args.limit)
        params['limit'] = parsed_args.limit
        params['marker'] = parsed_args.marker

        if parsed_args.address is not None:
            params['address'] = parsed_args.address
        if parsed_args.node is not None:
            params['node'] = parsed_args.node
        if parsed_args.portgroup is not None:
            params['portgroup'] = parsed_args.portgroup

        if parsed_args.detail:
            params['detail'] = parsed_args.detail
            columns = res_fields.PORT_DETAILED_RESOURCE.fields

        elif parsed_args.fields:
            params['detail'] = False
            fields = itertools.chain.from_iterable(parsed_args.fields)
            resource = res_fields.Resource(list(fields))
            columns = resource.fields
            params['fields'] = columns

        self.log.debug("params(%s)", params)
        data = client.port.list(**params)

        data = oscutils.sort_items(data, parsed_args.sort)

        return (columns,
                (oscutils.get_item_properties(s, columns, formatters={
                    'extra': utils.HashColumn},) for s in data))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import builtins
from typing import Any

from ironicclient.common import base
from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc


class Runbook(base.Resource):
    def __repr__(self) -> str:
        return "<Runbook %s>" % self._info


class RunbookManager(base.CreateManager[Runbook]):
    resource_class: type[Runbook] = Runbook
    _creation_attributes: list[str] = [
        'description', 'extra', 'name', 'owner', 'public', 'steps',
        'traits', 'uuid',
    ]
    _resource_name: str = 'runbooks'

    def get(
        self,
        runbook_id: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource | None:
        return self._get(
            resource_id=runbook_id,
            fields=fields,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def delete(
        self,
        runbook_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        return self._delete(
            resource_id=runbook_id,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def update(
        self,
        runbook_id: str,
        patch: list[dict[str, Any]],
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource:
        return self._update(
            resource_id=runbook_id,
            patch=patch,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def list(
        self,
        limit: int | None = None,
        marker: str | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        detail: bool = False,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        project: str | None = None,
        public: bool | None = None,
    ) -> list[Runbook]:
        """Retrieve a list of runbooks.

        :param marker: Optional, the UUID of a deploy template, eg the last
                       template from a previous result set. Return the next
                       result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of runbooks to return.
            2) limit == 0, return the entire list of runbooks.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about runbooks.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of runbooks.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(
                _("Can't fetch a subset of fields "
                  "with 'detail' set"))

        filters = utils.common_filters(
            marker, limit, sort_key, sort_dir,
            fields, detail=detail, public=public,
            project=project)
        path = ''
        if filters:
            path += '?' + '&'.join(filters)
        if limit is None:
            return self._list(
                self._path(path),
                "runbooks",
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path),
                "runbooks",
                limit=limit,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )

    def get_traits(
        self,
        runbook_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> builtins.list[str]:
        """Get traits for a runbook.

        Requires API version 1.112 or later.

        :param runbook_ident: runbook UUID or name.
        :param os_ironic_api_version: String version (e.g. "1.112") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        version_to_check = (os_ironic_api_version
                            or self.api.os_ironic_api_version)
        if version_to_check and not utils.check_api_version_support(
                version_to_check, "1.112"):
            raise exc.UnsupportedVersion(
                _("Runbook trait operations require API version 1.112 or "
                  "later."))

        path = f"{runbook_ident}/traits"
        return self._list_primitives(
            self._path(path), 'traits',
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)

    def add_trait(
        self,
        runbook_ident: str,
        trait: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> Runbook | None:
        """Add a trait to a runbook.

        Requires API version 1.112 or later.

        :param runbook_ident: runbook UUID or name.
        :param trait: trait to add to the runbook.
        :param os_ironic_api_version: String version (e.g. "1.112") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        version_to_check = (os_ironic_api_version
                            or self.api.os_ironic_api_version)
        if version_to_check and not utils.check_api_version_support(
                version_to_check, "1.112"):
            raise exc.UnsupportedVersion(
                _("Runbook trait operations require API version 1.112 or "
                  "later."))

        path = "%s/traits/%s" % (runbook_ident, trait)
        return self._update(
            resource_id=path,
            patch=None,
            method='PUT',
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)

    def remove_trait(
        self,
        runbook_ident: str,
        trait: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Remove a trait from a runbook.

        Requires API version 1.112 or later.

        :param runbook_ident: runbook UUID or name.
        :param trait: trait to remove from the runbook.
        :param os_ironic_api_version: String version (e.g. "1.112") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        version_to_check = (os_ironic_api_version
                            or self.api.os_ironic_api_version)
        if version_to_check and not utils.check_api_version_support(
                version_to_check, "1.112"):
            raise exc.UnsupportedVersion(
                _("Runbook trait operations require API version 1.112 or "
                  "later."))

        path = "%s/traits/%s" % (runbook_ident, trait)
        self._delete(
            resource_id=path,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)

    def remove_all_traits(
        self,
        runbook_ident: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Remove all traits from a runbook.

        Requires API version 1.112 or later.

        :param runbook_ident: runbook UUID or name.
        :param os_ironic_api_version: String version (e.g. "1.112") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        version_to_check = (os_ironic_api_version
                            or self.api.os_ironic_api_version)
        if version_to_check and not utils.check_api_version_support(
                version_to_check, "1.112"):
            raise exc.UnsupportedVersion(
                _("Runbook trait operations require API version 1.112 or "
                  "later."))

        path = "%s/traits" % runbook_ident
        self._delete(
            resource_id=path,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id)



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import logging
from typing import Any, cast

from ironicclient.common import filecache
from ironicclient.common import http
from ironicclient.common.http import DEFAULT_VER
from ironicclient.v1 import allocation
from ironicclient.v1 import chassis
from ironicclient.v1 import conductor
from ironicclient.v1 import deploy_template
from ironicclient.v1 import driver
from ironicclient.v1 import events
from ironicclient.v1 import inspection_rule
from ironicclient.v1 import node
from ironicclient.v1 import port
from ironicclient.v1 import portgroup
from ironicclient.v1 import runbook
from ironicclient.v1 import shard
from ironicclient.v1 import volume_connector
from ironicclient.v1 import volume_target

LOG: logging.Logger = logging.getLogger(__name__)


class Client(object):
    """Client for the Ironic v1 API.

    :param string endpoint_override: A user-supplied endpoint URL for the
                                     ironic service.
    :param session: A keystoneauth Session object (must be provided as
        a keyword argument).
    """

    def __init__(
        self,
        endpoint_override: str | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize a new client for the Ironic v1 API."""
        if not args and not kwargs.get('session'):
            raise TypeError("A session is required for creating a client, "
                            "use ironicclient.client.get_client to create "
                            "it automatically")

        allow_downgrade = kwargs.pop('allow_api_version_downgrade', False)
        if kwargs.get('os_ironic_api_version'):
            # TODO(TheJulia): We should sanity check os_ironic_api_version
            # against our maximum supported version, so the client fails
            # immediately upon an unsupported version being provided.
            # This logic should also likely live in common/http.py
            if allow_downgrade:
                if kwargs['os_ironic_api_version'] == 'latest':
                    raise ValueError(
                        "Invalid configuration defined. "
                        "The os_ironic_api_version can not be set "
                        "to 'latest' while allow_api_version_downgrade "
                        "is set.")
                # NOTE(dtantsur): here we allow the HTTP client to negotiate a
                # lower version if the requested is too high
                kwargs['api_version_select_state'] = "default"
            else:
                kwargs['api_version_select_state'] = "user"
        else:
            if endpoint_override:
                # If the user didn't specify a version, use a cached version if
                # one has been stored
                host, netport = http.get_server(endpoint_override)
                saved_version = filecache.retrieve_data(
                    host=cast(str, host),
                    port=cast(str, netport))
                if saved_version:
                    kwargs['api_version_select_state'] = "cached"
                    kwargs['os_ironic_api_version'] = saved_version
                else:
                    kwargs['api_version_select_state'] = "default"
                    kwargs['os_ironic_api_version'] = DEFAULT_VER
            else:
                LOG.debug('Cannot use cached API version since endpoint '
                          'override is not provided. Will negotiate again.')
                kwargs['api_version_select_state'] = "default"
                kwargs['os_ironic_api_version'] = DEFAULT_VER

        if endpoint_override:
            kwargs['endpoint_override'] = endpoint_override
        self.http_client = http._construct_http_client(*args, **kwargs)

        self.chassis = chassis.ChassisManager(self.http_client)
        self.node = node.NodeManager(self.http_client)
        self.port = port.PortManager(self.http_client)
        self.volume_connector = volume_connector.VolumeConnectorManager(
            self.http_client)
        self.volume_target = volume_target.VolumeTargetManager(
            self.http_client)
        self.driver = driver.DriverManager(self.http_client)
        self.runbook = runbook.RunbookManager(self.http_client)
        self.portgroup = portgroup.PortgroupManager(self.http_client)
        self.conductor = conductor.ConductorManager(self.http_client)
        self.events = events.EventManager(self.http_client)
        self.allocation = allocation.AllocationManager(self.http_client)
        self.deploy_template = deploy_template.DeployTemplateManager(
            self.http_client)
        self.shard = shard.ShardManager(self.http_client)
        self.inspection_rule = inspection_rule.InspectionRuleManager(
            self.http_client)

    @property
    def current_api_version(self) -> str | list[str]:
        """Return the current API version in use.

        This returns the version of the REST API that the API client
        is presently set to request. This value may change as a result
        of API version negotiation.
        """
        return self.http_client.os_ironic_api_version

    @property
    def is_api_version_negotiated(self) -> bool:
        """Returns True if microversion negotiation has occurred."""
        return self.http_client.api_version_select_state == 'negotiated'

    def negotiate_api_version(self) -> str:
        """Triggers negotiation with the remote API endpoint.

        :returns: the negotiated API version.
        """
        return self.http_client.negotiate_version(
            self.http_client.session, None)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import copy

import testtools
from testtools.matchers import HasLength

from ironicclient import exc
from ironicclient.tests.unit import utils
import ironicclient.v1.runbook

RUNBOOK = {'uuid': '11111111-2222-3333-4444-555555555555',
           'name': 'CUSTOM_RUNBOOK',
           'steps': {},
           'extra': {}}

RUNBOOK2 = {'uuid': '55555555-4444-3333-2222-111111111111',
            'name': 'CUSTOM_RUNBOOK2',
            'steps': {},
            'extra': {}}

CREATE_RUNBOOK = copy.deepcopy(RUNBOOK)
del CREATE_RUNBOOK['uuid']

CREATE_RUNBOOK_WITH_UUID = copy.deepcopy(RUNBOOK)

UPDATED_RUNBOOK = copy.deepcopy(RUNBOOK)
NEW_NAME = 'CUSTOM_RUNBOOK3'
UPDATED_RUNBOOK['name'] = NEW_NAME

fake_responses = {
    '/v1/runbooks':
    {
        'GET': (
            {},
            {"runbooks": [RUNBOOK]},
        ),
        'POST': (
            {},
            CREATE_RUNBOOK,
        ),
    },
    '/v1/runbooks/?detail=True':
    {
        'GET': (
            {},
            {"runbooks": [RUNBOOK]},
        ),
    },
    '/v1/runbooks/?fields=uuid,name':
    {
        'GET': (
            {},
            {"runbooks": [RUNBOOK]},
        ),
    },
    '/v1/runbooks/%s' % RUNBOOK['uuid']:
    {
        'GET': (
            {},
            RUNBOOK,
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_RUNBOOK,
        ),
    },
    '/v1/runbooks/%s?fields=uuid,name' % RUNBOOK['uuid']:
    {
        'GET': (
            {},
            RUNBOOK,
        ),
    },
    '/v1/runbooks/%s/traits' % RUNBOOK['uuid']:
    {
        'GET': (
            {},
            {'traits': ['CUSTOM_FOO', 'CUSTOM_BAR']},
        ),
        'DELETE': (
            {},
            None,
        ),
    },
    '/v1/runbooks/%s/traits/CUSTOM_FOO' % RUNBOOK['uuid']:
    {
        'PUT': (
            {},
            None,
        ),
        'DELETE': (
            {},
            None,
        ),
    },
}

fake_responses_pagination = {
    '/v1/runbooks':
    {
        'GET': (
            {},
            {"runbooks": [RUNBOOK],
             "next": "http://127.0.0.1:6385/v1/runbooks/?limit=1"}
        ),
    },
    '/v1/runbooks/?limit=1':
    {
        'GET': (
            {},
            {"runbooks": [RUNBOOK2]}
        ),
    },
    '/v1/runbooks/?marker=%s' % RUNBOOK['uuid']:
    {
        'GET': (
            {},
            {"runbooks": [RUNBOOK2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/runbooks/?sort_key=updated_at':
    {
        'GET': (
            {},
            {"runbooks": [RUNBOOK2, RUNBOOK]}
        ),
    },
    '/v1/runbooks/?sort_dir=desc':
    {
        'GET': (
            {},
            {"runbooks": [RUNBOOK2, RUNBOOK]}
        ),
    },
}


class RunbookManagerTest(testtools.TestCase):

    def setUp(self) -> None:
        super(RunbookManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        # Set API version to support trait operations
        self.api.os_ironic_api_version = "1.112"
        self.mgr = ironicclient.v1.runbook.RunbookManager(
            self.api)

    def test_runbooks_list(self) -> None:
        runbooks = self.mgr.list()
        expect = [
            ('GET', '/v1/runbooks', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(runbooks))

    def test_runbooks_list_detail(self) -> None:
        runbooks = self.mgr.list(detail=True)
        expect = [
            ('GET', '/v1/runbooks/?detail=True', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(runbooks))

    def test_runbook_list_fields(self) -> None:
        runbooks = self.mgr.list(fields=['uuid', 'name'])
        expect = [
            ('GET', '/v1/runbooks/?fields=uuid,name', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(runbooks))

    def test_runbook_list_detail_and_fields_fail(self) -> None:
        self.assertRaises(exc.InvalidAttribute, self.mgr.list,
                          detail=True, fields=['uuid', 'name'])

    def test_runbooks_list_limit(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.runbook.RunbookManager(
            self.api)
        runbooks = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/runbooks/?limit=1', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(runbooks, HasLength(1))

    def test_runbooks_list_marker(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.runbook.RunbookManager(
            self.api)
        runbooks = self.mgr.list(marker=RUNBOOK['uuid'])
        expect = [
            ('GET',
             '/v1/runbooks/?marker=%s' % RUNBOOK['uuid'], {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(runbooks, HasLength(1))

    def test_runbooks_list_pagination_no_limit(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.runbook.RunbookManager(
            self.api)
        runbooks = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/runbooks', {}, None),
            ('GET', '/v1/runbooks/?limit=1', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(runbooks, HasLength(2))

    def test_runbooks_list_sort_key(self) -> None:
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = ironicclient.v1.runbook.RunbookManager(
            self.api)
        runbooks = self.mgr.list(sort_key='updated_at')
        expect = [
            ('GET', '/v1/runbooks/?sort_key=updated_at', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(runbooks))

    def test_runbooks_list_sort_dir(self) -> None:
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = ironicclient.v1.runbook.RunbookManager(
            self.api)
        runbooks = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/runbooks/?sort_dir=desc', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(runbooks))

    def test_runbooks_show(self) -> None:
        runbook = self.mgr.get(RUNBOOK['uuid'])
        expect = [
            ('GET', '/v1/runbooks/%s' % RUNBOOK['uuid'], {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(RUNBOOK['uuid'], runbook.uuid)
        self.assertEqual(RUNBOOK['name'], runbook.name)
        self.assertEqual(RUNBOOK['steps'], runbook.steps)
        self.assertEqual(RUNBOOK['extra'], runbook.extra)

    def test_runbook_show_fields(self) -> None:
        runbook = self.mgr.get(RUNBOOK['uuid'],
                               fields=['uuid', 'name'])
        expect = [
            ('GET', '/v1/runbooks/%s?fields=uuid,name' %
             RUNBOOK['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(RUNBOOK['uuid'], runbook.uuid)
        self.assertEqual(RUNBOOK['name'], runbook.name)

    def test_create(self) -> None:
        runbook = self.mgr.create(**CREATE_RUNBOOK)
        expect = [
            ('POST', '/v1/runbooks', {}, CREATE_RUNBOOK),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(runbook)

    def test_create_with_uuid(self) -> None:
        runbook = self.mgr.create(**CREATE_RUNBOOK_WITH_UUID)
        expect = [
            ('POST', '/v1/runbooks', {},
             CREATE_RUNBOOK_WITH_UUID),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(runbook)

    def test_delete(self) -> None:
        self.mgr.delete(runbook_id=RUNBOOK['uuid'])
        expect = [
            ('DELETE', '/v1/runbooks/%s' % RUNBOOK['uuid'], {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)

    def test_update(self) -> None:
        patch = {'op': 'replace',
                 'value': NEW_NAME,
                 'path': '/name'}
        runbook = self.mgr.update(
            runbook_id=RUNBOOK['uuid'], patch=patch)
        expect = [
            ('PATCH', '/v1/runbooks/%s' % RUNBOOK['uuid'],
             {}, patch),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(NEW_NAME, runbook.name)

    def test_runbook_get_traits(self) -> None:
        traits = self.mgr.get_traits(RUNBOOK['uuid'])
        expect = [
            ('GET', '/v1/runbooks/%s/traits' % RUNBOOK['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(['CUSTOM_FOO', 'CUSTOM_BAR'], traits)

    def test_runbook_add_trait(self) -> None:
        trait = 'CUSTOM_FOO'
        resp = self.mgr.add_trait(RUNBOOK['uuid'], trait)
        expect = [
            ('PUT', '/v1/runbooks/%s/traits/%s' % (RUNBOOK['uuid'], trait),
                {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(resp)

    def test_runbook_remove_all_traits(self) -> None:
        resp = self.mgr.remove_all_traits(RUNBOOK['uuid'])
        expect = [
            ('DELETE', '/v1/runbooks/%s/traits' % RUNBOOK['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(resp)

    def test_runbook_remove_trait(self) -> None:
        trait = 'CUSTOM_FOO'
        resp = self.mgr.remove_trait(RUNBOOK['uuid'], trait)
        expect = [
            ('DELETE', '/v1/runbooks/%s/traits/%s' % (RUNBOOK['uuid'], trait),
                {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(resp)

    def test_runbook_get_traits_old_api_version(self) -> None:
        # Test with API version that doesn't support traits
        self.api.os_ironic_api_version = "1.111"
        from ironicclient.common.apiclient.exceptions import UnsupportedVersion
        self.assertRaises(UnsupportedVersion, self.mgr.get_traits,
                          RUNBOOK['uuid'])

    def test_runbook_add_trait_old_api_version(self) -> None:
        # Test with API version that doesn't support traits
        self.api.os_ironic_api_version = "1.111"
        trait = 'CUSTOM_FOO'
        from ironicclient.common.apiclient.exceptions import UnsupportedVersion
        self.assertRaises(UnsupportedVersion, self.mgr.add_trait,
                          RUNBOOK['uuid'], trait)

    def test_runbook_remove_trait_old_api_version(self) -> None:
        # Test with API version that doesn't support traits
        self.api.os_ironic_api_version = "1.111"
        trait = 'CUSTOM_FOO'
        from ironicclient.common.apiclient.exceptions import UnsupportedVersion
        self.assertRaises(UnsupportedVersion, self.mgr.remove_trait,
                          RUNBOOK['uuid'], trait)

    def test_runbook_remove_all_traits_old_api_version(self) -> None:
        # Test with API version that doesn't support traits
        self.api.os_ironic_api_version = "1.111"
        from ironicclient.common.apiclient.exceptions import UnsupportedVersion
        self.assertRaises(UnsupportedVersion, self.mgr.remove_all_traits,
                          RUNBOOK['uuid'])



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from unittest import mock

import testtools
from testtools import matchers

from ironicclient import exc
from ironicclient.tests.unit import utils
from ironicclient.v1 import driver


DRIVER1 = {
    'name': 'fake',
    'type': 'dynamic',
    'hosts': ['fake-host1', 'fake-host2'],
    'default_bios_interface': 'bios',
    'default_boot_interface': 'boot',
    'default_console_interface': 'console',
    'default_deploy_interface': 'deploy',
    'default_firmware_interface': 'firmware',
    'default_inspect_interface': 'inspect',
    'default_management_interface': 'management',
    'default_network_interface': 'network',
    'default_power_interface': 'power',
    'default_raid_interface': 'raid',
    'default_vendor_interface': 'vendor',
    'enabled_bios_interfaces': ['bios', 'bios2'],
    'enabled_boot_interfaces': ['boot', 'boot2'],
    'enabled_console_interfaces': ['console', 'console2'],
    'enabled_deploy_interfaces': ['deploy', 'deploy2'],
    'enabled_firmware_interfaces': ['firmware', 'firmware2'],
    'enabled_inspect_interfaces': ['inspect', 'inspect2'],
    'enabled_management_interfaces': ['management', 'management2'],
    'enabled_network_interfaces': ['network', 'network2'],
    'enabled_power_interfaces': ['power', 'power2'],
    'enabled_raid_interfaces': ['raid', 'raid2'],
    'enabled_vendor_interfaces': ['vendor', 'vendor2'],
}
DRIVER2 = {
    'name': 'pxe_ipminative',
    'type': 'classic',
    'hosts': ['fake-host1', 'fake-host2'],
}

DRIVER2_PROPERTIES = {
    "username": "username. Required.",
    "password": "password. Optional.",
    "address": "IP of the node. Required.",
}

DRIVER_VENDOR_PASSTHRU_METHOD = {"lookup": {"attach": "false",
                                            "http_methods": ["POST"],
                                            "description": "",
                                            "async": "false"}}
DRIVER2_RAID_LOGICAL_DISK_PROPERTIES = {
    "property1": "description1",
    "property2": "description2",
}

fake_responses = {
    '/v1/drivers':
    {
        'GET': (
            {},
            {'drivers': [DRIVER1]},
        ),
    },
    '/v1/drivers/?fields=name,hosts':
    {
        'GET': (
            {},
            {'drivers': [DRIVER1]}
        )
    },
    '/v1/drivers/%s' % DRIVER1['name']:
    {
        'GET': (
            {},
            DRIVER1
        ),
    },
    '/v1/drivers/%s?fields=name,hosts' % DRIVER1['name']:
    {
        'GET': (
            {},
            DRIVER1,
        ),
    },
    '/v1/drivers/%s/properties' % DRIVER2['name']:
    {
        'GET': (
            {},
            DRIVER2_PROPERTIES,
        ),
    },
    '/v1/drivers/%s/vendor_passthru/methods' % DRIVER1['name']:
    {
        'GET': (
            {},
            DRIVER_VENDOR_PASSTHRU_METHOD,
        ),
    },
    '/v1/drivers/%s/raid/logical_disk_properties' % DRIVER2['name']:
    {
        'GET': (
            {},
            DRIVER2_RAID_LOGICAL_DISK_PROPERTIES,
        ),
    },
}


class DriverManagerTest(testtools.TestCase):

    def setUp(self) -> None:
        super(DriverManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = driver.DriverManager(self.api)

    def test_driver_list(self) -> None:
        drivers = self.mgr.list()
        expect = [
            ('GET', '/v1/drivers', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(drivers, matchers.HasLength(1))

    def test_driver_show(self) -> None:
        driver_ = self.mgr.get(DRIVER1['name'])
        expect = [
            ('GET', '/v1/drivers/%s' % DRIVER1['name'], {}, None)
        ]
        self.assertEqual(expect, self.api.calls)

        driver_attr = {}
        for attr in DRIVER1.keys():
            driver_attr[attr] = getattr(driver_, attr)

        self.assertEqual(DRIVER1, driver_attr)

    def test_driver_list_fields(self) -> None:
        drivers = self.mgr.list(fields=['name', 'hosts'])
        expect = [
            ('GET', '/v1/drivers/?fields=name,hosts', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(drivers, matchers.HasLength(1))

    def test_driver_show_fields(self) -> None:
        driver_ = self.mgr.get(DRIVER1['name'], fields=['name', 'hosts'])
        expect = [
            ('GET', '/v1/drivers/%s?fields=name,hosts' %
             DRIVER1['name'], {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(DRIVER1['name'], driver_.name)
        self.assertEqual(DRIVER1['hosts'], driver_.hosts)

    def test_driver_properties(self) -> None:
        properties = self.mgr.properties(DRIVER2['name'])
        expect = [
            ('GET', '/v1/drivers/%s/properties' % DRIVER2['name'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(DRIVER2_PROPERTIES, properties)

    def test_driver_raid_logical_disk_properties(self) -> None:
        properties = self.mgr.raid_logical_disk_properties(DRIVER2['name'])
        expect = [
            ('GET',
             '/v1/drivers/%s/raid/logical_disk_properties' % DRIVER2['name'],
             {}, None)]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(DRIVER2_RAID_LOGICAL_DISK_PROPERTIES, properties)

    @mock.patch.object(driver.DriverManager, '_list', autospec=True)
    def test_driver_raid_logical_disk_properties_indexerror(
        self, _list_mock: mock.MagicMock,
    ) -> None:
        _list_mock.side_effect = IndexError

        properties = self.mgr.raid_logical_disk_properties(DRIVER2['name'])

        _list_mock.assert_called_once_with(
            self.mgr,
            '/v1/drivers/%s/raid/logical_disk_properties' % DRIVER2['name'],
            os_ironic_api_version=None, global_request_id=None)
        self.assertEqual({}, properties)

    @mock.patch.object(driver.DriverManager, 'update', autospec=True)
    def test_vendor_passthru_update(
        self, update_mock: mock.MagicMock,
    ) -> None:
        # For now just mock the tests because vendor-passthru doesn't return
        # anything to verify.
        vendor_passthru_args = {'arg1': 'val1'}
        kwargs = {
            'driver_name': 'driver_name',
            'method': 'method',
            'args': vendor_passthru_args
        }

        final_path = 'driver_name/vendor_passthru/method'
        for http_method in ('POST', 'PUT', 'PATCH'):
            kwargs['http_method'] = http_method
            self.mgr.vendor_passthru(**kwargs)
            update_mock.assert_called_once_with(
                mock.ANY, final_path, vendor_passthru_args,
                http_method=http_method, os_ironic_api_version=None,
                global_request_id=None)
            update_mock.reset_mock()

    @mock.patch.object(driver.DriverManager, 'get', autospec=True)
    def test_vendor_passthru_get(
        self, get_mock: mock.MagicMock,
    ) -> None:
        kwargs = {
            'driver_name': 'driver_name',
            'method': 'method',
            'http_method': 'GET',
        }

        final_path = 'driver_name/vendor_passthru/method'
        self.mgr.vendor_passthru(**kwargs)
        get_mock.assert_called_once_with(
            mock.ANY, final_path, os_ironic_api_version=None,
            global_request_id=None)

    @mock.patch.object(driver.DriverManager, 'delete', autospec=True)
    def test_vendor_passthru_delete(
        self, delete_mock: mock.MagicMock,
    ) -> None:
        kwargs = {
            'driver_name': 'driver_name',
            'method': 'method',
            'http_method': 'DELETE',
        }

        final_path = 'driver_name/vendor_passthru/method'
        self.mgr.vendor_passthru(**kwargs)
        delete_mock.assert_called_once_with(
            mock.ANY, final_path, os_ironic_api_version=None,
            global_request_id=None)

    @mock.patch.object(driver.DriverManager, 'delete', autospec=True)
    def test_vendor_passthru_unknown_http_method(
        self, delete_mock: mock.MagicMock,
    ) -> None:
        kwargs = {
            'driver_name': 'driver_name',
            'method': 'method',
            'http_method': 'UNKNOWN',
        }
        self.assertRaises(exc.InvalidAttribute, self.mgr.vendor_passthru,
                          **kwargs)

    def test_vendor_passthru_methods(self) -> None:
        vendor_methods = self.mgr.get_vendor_passthru_methods(DRIVER1['name'])
        expect = [
            ('GET', '/v1/drivers/%s/vendor_passthru/methods' % DRIVER1['name'],
             {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(DRIVER_VENDOR_PASSTHRU_METHOD, vendor_methods)
